
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef WIN32
    #include <winsock2.h>
    #include  <ws2tcpip.h>
#else
    #include <sys/socket.h>
    #include <sys/time.h>
    #include <netinet/in.h>
    #include <netdb.h>
#endif
#include <fcntl.h>


#include "uSQLcClient.h"

struct linger lg;
socklen_t lgs;
void SockNoBlock(int sock){
int flgs;
unsigned long nonblock=1;

#ifdef WIN32
    // set socket non-blocking
    ioctlsocket(sock,FIONBIO,&nonblock);
#else
    // Set socket non-blocking
    flgs=fcntl(sock,F_GETFL,0);
    flgs|=O_NONBLOCK;
    fcntl(sock,F_SETFL,flgs);
#endif

    // set linger
    lg.l_onoff=0;
    lg.l_linger=0;
    lgs=sizeof(lg);
    setsockopt(sock,SOL_SOCKET,SO_LINGER,(const char*)&lg,lgs);
}

int SetName(struct sockaddr_in *name,char *server,int port){
struct hostent *lh;
int b1,b2,b3,b4;

    //First see if server appears to be a dotted IP:
    if(sscanf(server,"%i.%i.%i.%i",b1,b2,b3,b4)==4)
    	name->sin_addr.s_addr=htonl((b1*0x1000000)+(b2*0x10000)+(b3*0x100)+b4);
    else{
        if(!(lh=gethostbyname(server)))goto abort;
        if((lh->h_addrtype!=AF_INET)||(lh->h_length!=4))goto abort;
		name->sin_addr=*(struct in_addr *) lh->h_addr;
    }

    name->sin_family = AF_INET;
    name->sin_port = htons (port);

    return 0;
abort:
    return 1;
}

// Allthougth a local function, wait RX is where the work gets done
// and defines the overall behaviour

// Queries sent to the server always finish with ':OK'. waitRX
// Will eat input unti the :OK arrives.
// If the return includes record data it will call cback each
// time a record is recieved (if cback is NULL then data is trashed.
// If an Error record arrives then the eback will be invoked (providing it
// is non null.
// waitRX will also intercept any #PPRAGAM LEVEL n replies and update the
// seclevel value in udb accordingly.
// If cback returns non zero then it will be set to NULL (there is currently
// no mechnism to stop data arriving from the server, it must be trashed, but
// note that servers have a global clamp on reply sizes.
// waitRX will timeout on the current timeout value, allthougth this
// only takes into account the 'waiting' time and not working time. This means
// that the timeouts will not prevent stalling by the callback functions.
//
// Returns 0 if all OK, uSQLCerrTEXEC on a timeout, and a
// uSQLCerrERR if an err record has been recieved during the execution.
// The incoming buffer is reference in the udb structure, so callbacks may snoop it
// (Headers and Errors are left in the buffer, records are overwritten).
int waitRX(uSQLCdb *udb,void *tag,DatCallBack cback,ErrCallBack eback){

int i,timecnt,rxed,retcode=0,rbufsize,incsz=256,minsz=64;
char *rbuff=NULL,*cur,*line_base,*pcur,*temp;
char tcr='\r',tlf='\n',foo;
int bufferdone;
char **heads=NULL; //Pointers to column header strings
char **fields=NULL;
int *fsizes=NULL;
int headscnt=0;     //Number of headers in heads
int headnum,headlen,headoff;
size_t mod;
int RecCount=0;
int ColCur;
struct timeval stimes;
int slct=0;
#ifndef WIN32
struct timespec stime,ttime;
#endif



    retcode=0;
    timecnt=udb->timeout;

    // Set up initial buffer (queries which don't return rs's are small)
    // While handling rs's this will grow to size of biggest record.
    if(!(rbuff=malloc(256)))return uSQLCerrMALLOC;
    rbufsize=256;
    pcur=line_base=cur=rbuff;

    while(1){
        // Make sure we have enougth space to do a read
        if((rbufsize-(cur-rbuff))<minsz){
        	rbufsize+=incsz;
        	temp=rbuff;
        	if(!(rbuff=realloc(rbuff,rbufsize))){
        		retcode= uSQLCerrMALLOC;
                goto abort;
        	}
        	// If realloc returns a new base, we must modify any
        	// existing pointers in fields
        	if((mod=(rbuff-temp))!=0){
        		line_base+=mod;
        		cur+=mod;
                if(RecCount!=0)
                    for(i=0;i<headscnt;i++)if(fields[i])fields[i]+=mod;
        	}
        }

        // read data
        rxed=recv(udb->sock,cur,rbufsize-(cur-rbuff)-1,0);
        if(rxed<0)rxed=0; // Error codes to no rx
        cur+=rxed;

        // See if we have a line. A line terminator is considered
        // to be a '0', a '\r' a '\n' or ETX '\0x03'
        // When an ETX has been recieved \r and \n are no longer valid.

        // If line_base is pointing at a line terminator char, move it
        // up until it isn't. (exceptional case where >1 terminator rx'ed)
        while((line_base<cur)&&((*line_base==0)||(*line_base==3)||(*line_base==tcr)||(*line_base==tlf))){
            if(*line_base==3)tcr=tlf=0;
            line_base++;
        }

        // pcur searches from line_base to cur to find a line break
        for(pcur=line_base;pcur<cur;pcur++)
            if((*pcur==0)||(*pcur==3)||(*pcur==tcr)||(*pcur==tlf)){
            	if(*pcur==3)tcr=tlf=0;
            	*pcur=0;    // Force a null terminator
            	break;
            }

        // ok, if there is a line break, elaborate it
        if((pcur>line_base)&&(pcur<cur)&&(*pcur==0)){
        	// bufferdone set non zero if handler adjusted buffer
        	bufferdone=0;

            // Check for records (no logical xor operator?)
            if((line_base[0]!=':')||((line_base[0]==':')&&(line_base[1]=='F'))){
                // Ignore unformatted lines before record section
	           	if(!RecCount){
	           		line_base=pcur;
	           		continue;
            	}
            	//Handle a null field
            	if(*line_base=='!'){
                    fields[ColCur]=NULL;
                    fsizes[ColCur]=0;
            	}
                //Handle a formal field
                else if(sscanf(line_base,":F%i %n",&headlen,&headoff)==1){
                	if(headlen){
                        fields[ColCur]=line_base+headoff;
                        fsizes[ColCur]=headlen;
                	}
                	else{
                        fields[ColCur]=NULL;
                        fsizes[ColCur]=0;
                	}
                }
                // Else an informal field
                else{
                	fields[ColCur]=line_base;
                	fsizes[ColCur]=strlen(line_base);
                }

                // If we have got all the columns, invoke the call back,
                // Otherwise just shift up line_base so previous fields are
                // left in place
                if((++ColCur)<headscnt){
                	line_base=pcur;
                	bufferdone=1;
                }
                else{
                	if(cback)if(!cback(tag,headscnt,fields,heads,fsizes))cback=NULL;
                	for(i=0;i<headscnt;i++)fields[i]=NULL;  // init to Null
                    ColCur=0;
                    // Move back buffer
                    memmove(rbuff,pcur,cur-pcur);
                    cur=rbuff+(cur-pcur);
                    line_base=pcur=rbuff;
                    bufferdone=1;
                }
            }

            // Check for an error record
            else if(!strncmp(line_base,":Err",4)){
            	if(eback)eback(line_base);
            	retcode=uSQLCerrERR;
            }

            // Check for a column header
            else if((sscanf(line_base,":H%i:%i %n",&headnum,&headlen,&headoff))==2){
            	//Check that the header has arrived in the correct sequence
            	if((headnum!=headscnt+1)||RecCount){
            		retcode=uSQLCerrFORM;
            		goto abort;
            	}
            	// Sanity check the size before mallocing!
            	if((headlen<1)||(headlen>2048)){    // Fascist programming! (my limits)
                    retcode=uSQLCerrHSZ;
                    goto abort;
            	}
            	// Increase the size of heads
            	// !!! Passing NULL to realloc may not be portable
            	if(!(heads=(char**)realloc(heads,(headscnt+1)*sizeof(char*)))){
            		retcode=uSQLCerrMALLOC;
                    goto abort;
            	}
            	heads[headscnt]=NULL;
            	headscnt+=1;    // We must do this straight away to ensure free() on abort
            	// Make space for the header and put it
                if(!(heads[headscnt-1]=(char*)malloc(headlen+1))){
            		retcode=uSQLCerrMALLOC;
                    goto abort;
            	}
            	strncpy(heads[headscnt-1],line_base+headoff,headlen);
            	*(heads[headscnt-1]+headlen)=0;
            }

            // Check for the head/record switch
            else if(!strncmp(line_base,":R",2)){

                // Check we have at least 1 header and block additional
                // headers by setting RecCount=1
                if(!headscnt){
            		retcode=uSQLCerrFORM;
            		goto abort;
            	}
                RecCount=1;
                //Increase these parameters if we are to be handling records
                incsz=512;
                minsz=128;
                // Allocate space for field pointers
                if(!(fields=(char**)malloc(headscnt*sizeof(char*)))){
             		retcode=uSQLCerrMALLOC;
                    goto abort;
            	}
                // Allocate space for field sizes
                if(!(fsizes=(int*)malloc(headscnt*sizeof(int)))){
             		retcode=uSQLCerrMALLOC;
                    goto abort;
            	}

               	for(i=0;i<headscnt;i++)fields[i]=NULL;  // init to Null
               	ColCur=0;
            }

            // Check for OK
            else if(!strncmp(line_base,":OK",3))
            	goto abort; // (if nothing untoward retcode=no error)


        	// Check for change of seclevel
        	else if((sscanf(line_base,":PPRAGMA USELEVEL %i",&udb->seclevel))==1);

            // Check for client ID
            else sscanf(line_base,":PPRAGMA ID %i",&udb->clientID);


            if(!bufferdone)line_base=pcur;
        }


        // timeouts and wait for activity
        if(timecnt<=0){
        	retcode=uSQLCerrTEXEC;
        	goto abort;
        }

        // Something wrong with the select here,
       // Will poll till I get to the
        // bottom of this
/*        #ifdef WIN32
        while(timecnt){
            stimes.tv_sec=0;
            stimes.tv_usec=5000;
            FD_ZERO(&udb->srefs);
            FD_SET(udb->sock,&udb->srefs);
            slct=select(FD_SETSIZE,&udb->srefs,NULL,NULL,&stimes);
            if(slct==0)timecnt-=5;
            else break;
        }
        #else  */
        if((rxed)){
            #ifdef WIN32
            Sleep(3);
            #else
            stime.tv_sec=0;
            stime.tv_nsec=3000000;
            nanosleep(&stime,&ttime);
            #endif
            timecnt -=3;
            slct=0;
        }
//        #endif

        /*#else
        if(!rxed){
            stime.tv_sec=0;
            stime.tv_nsec=5000000;
            //nanosleep(&stime,&ttime);
            //timecnt-=50;
         }
         #endif */
    }

abort:
    free(fsizes);
    free(rbuff);
    free(fields);
    while(headscnt)free(heads[--headscnt]);
    free(heads);
    return retcode;
}

// will try sending data till timeout is reached
// if len is 0 then it will be set using strlen
// if len -1 then it is also missing the terminator
int sendTX(uSQLCdb *udb,char *data,int len){
int timecnt,txed;
#ifndef WIN32
struct timespec stime,ttime;
#endif

    if(!data)return 0;  //can be called with nothing to send
    timecnt=udb->timeout;
    if(!len)len=strlen(data);
    // if -1 then switch the trailing null to a terminator, as null no longer needed
    if(len<0){
    	len=strlen(data);
    	*(data+(len++))=3;
    }


    txed=0;
    while(1){
        txed+=send(udb->sock,data+txed,len-txed,0);

        if(!(len-txed))return uSQLCerrNONE;
        if(timecnt<=0)
            return uSQLCerrTEXEC;
        #ifdef WIN32
        Sleep(5);
        #else
        stime.tv_sec=0;
        stime.tv_nsec=5000000;
        nanosleep(&stime,&ttime);
        #endif
        timecnt-=5;
    }
}


// Opens up a connection to a uSQLdb server. On success returns a pointer
// to a uSQLdb structure. On failure returns NULL, ErrCode may have been
// set to give some idea for the failure reason.
uSQLCdb *uSQLCopen(char *user,char *pass,char *server,int port,int timeout,int *ErrCode){
uSQLCdb *udb;
int res;
struct sockaddr_in name;
char sends[100];

    // Make a new object or return NULL
    *ErrCode=uSQLCerrNONE;
    if(!(udb=(uSQLCdb*)malloc(sizeof(uSQLCdb)))){
    	*ErrCode=uSQLCerrMALLOC;
    	return NULL;
    }

    // Sanitize the new structure
    udb->sock=0;
    udb->timeout=timeout;
    udb->clientID=0;
    FD_ZERO(&udb->srefs);

    // open a socket
    if(!(udb->sock=(socket(PF_INET,SOCK_STREAM,0)))){
    	*ErrCode=uSQLCerrSOCK;
    	goto abort;
    }
    FD_SET(udb->sock,&udb->srefs);


    // Connection will be made using system timeouts
    if((SetName(&name,server,port))){
    	*ErrCode=uSQLCerrNAME;
    	goto abort;
    }

    if((connect(udb->sock,(struct sockaddr*)&name,(socklen_t)sizeof(name)))){
    	*ErrCode=uSQLCerrCON;
    	goto abort;
    }

    // Set sock nonblock to use timeouts
    SockNoBlock(udb->sock);


    //Check we are connected to uSQLite
    if(sendTX(udb,":PPRAGMA VER\r",0)){
    	*ErrCode=uSQLCerrRESP;
    	goto abort;
    }
    if(waitRX(udb,NULL,NULL,NULL)){
    	*ErrCode=uSQLCerrRESP;
    	goto abort;
    }

    //Switch to ETX
    if(sendTX(udb,":PPRAGMA ETX\r",0)){
    	*ErrCode=uSQLCerrRESP;
    	goto abort;
    }
    if(waitRX(udb,NULL,NULL,NULL)){
    	*ErrCode=uSQLCerrRESP;
    	goto abort;
    }

    //Send user name and pass
    sprintf(sends,":PPRAGMA USER %s\03",user);
    if(sendTX(udb,sends,0)){
    	*ErrCode=uSQLCerrRESP;
    	goto abort;
    }
    if(waitRX(udb,NULL,NULL,NULL)){
    	*ErrCode=uSQLCerrRESP;
    	goto abort;
    }
    sprintf(sends,":PPRAGMA PASS %s\03",pass);
    if(sendTX(udb,sends,0)){
    	*ErrCode=uSQLCerrRESP;
    	goto abort;
    }
    if(waitRX(udb,NULL,NULL,NULL)){
    	*ErrCode=uSQLCerrRESP;
    	goto abort;
    }

    // At this point we should be logged in.
    // The calling routine should check the seclevel
    // to see if the login got the expected results
    return udb;

abort:
    uSQLCclose(udb);
    return NULL;
}

// Closes a connection and frees up any memory allocated.
void uSQLCclose(uSQLCdb *udb){
    if(!udb)return;
    if(udb->sock)
        #ifdef WIN32
        closesocket(udb->sock);
        #else
        close(udb->sock);
        #endif
    free(udb);
}

// This has been included as a function so
// behaviour can be modified. Do not set udb->timeouts directly
void uSQLCtimeouts(uSQLCdb *udb,int timeouts){
	udb->timeout=timeouts;
}

int uSQLCexec(uSQLCdb *udb,char *sql,DatCallBack cback,void *tag,ErrCallBack eback){
int retval;

    retval=sendTX(udb,sql,-1);
    if(retval)return retval;

    return waitRX(udb,tag,cback,eback);

}
