#include <stdlib.h>
#include <stdio.h>

#include "Chink.h"

// Make a new empty Chink
Chink *ChinkMake(int maxsize){
Chink *ck;

    ck=malloc(sizeof(Chink));
    if(!ck)return ck;
    ck->ch=malloc(CHINKBLOK);
    if(!ck->ch){
        free(ck);
        return NULL;
    }
    
    ck->tr='\r';
    ck->tn='\n';
    ck->maxsize=maxsize;
    ck->len=0;
    ck->allocated=CHINKBLOK;
    
    return ck;
}

// Destroy a Chink
void ChinkDestroy(Chink *ck){
    if(ck->ch)free(ck->ch);
    free(ck);
}

// Clear contents
void ChinkReset(Chink *ck){
char *nch;

    if(nch=malloc(CHINKBLOK)){
     	free(ck->ch);
    	ck->ch=nch;
        ck->allocated=CHINKBLOK;
    }
    
    // If we could not alloc, try realloc
    if(nch=realloc(ck->ch,CHINKBLOK)){
        ck->ch=nch;
       	ck->allocated=CHINKBLOK;
    }

    // If we could not even realloc then allocated size
    // remains the same
    
    *(ck->ch)=0;
   	ck->len=0;
    	
}


// Add data, return new len or -1 if allocate error
// if len<0 will use strlen for length
int ChinkAdd(Chink *ck,char *data,int len){

	if(len<0)len=strlen(data);
	
    if(ChinkBuffFor(ck,len)<0)return -1;
    
    memcpy(ck->ch+ck->len,data,len);
    ck->len+=len;
			
}


// Make sure buffer can accept maxsize, returns extra space available or -1 for err
int ChinkBuffFor(Chink *ck,int maxsize){
int spc,al;
char *nch;
    
    if((spc=ck->allocated-maxsize)>=0)return spc;
    spc=abs(spc);
    
    // we allocate at least CHINKBLOK
    al=spc>CHINKBLOK?spc:CHINKBLOK;
    if(al>ck->maxsize)return -1;
    if(!(nch=realloc(ck->ch,al)))return-1;
    
    ck->allocated+=al;
    
    return al-spc;
}
    
    

// Returns char* to insert point
char *ChinkIns(Chink *ck){
    return ck->ch+ck->len;
}

// stripbreaks removes any leading line breaks/nulls. Should not
// normally be required so we are not bothered about efficiency
char *stripbreaks(Chink *ck){
int i=0;

    while(((ck->ch[i]==ck->tr)||(ck->ch[i]==ck->tn)||(ck->ch[i]==0))&&i<ck->len)i++;
    if(!i)return ck->ch;
    
    ck->len-=i;
    memmove(ck->ch,ck->ch+i,ck->len);
    return ck->ch;
}


	
// Return len of first line or 0 if no line break/null char (excludes break/null chars)
// ChinkBreak will also strip any leading line break/null chars
// The first break char will be forced to null so we have a c string
// The return value is the length of the line excluding the trailing NULL
int ChinkBreak(Chink *ck){
char *cur;
int i=0;

    cur=ck->ch;
    if((*cur==ck->tr)||(*cur==ck->tn)||(*cur==0))cur=stripbreaks(ck);
    for(cur;cur<ck->ch+ck->len;cur++){
        if((*cur==ck->tr)||(*cur==ck->tn)||(*cur==0)){
        	*cur=0;
            break;
        }        	
        i++;
    }
    return i<ck->len? i : 0;    
}
    

// Remake removing consume chars at start, returns len.
// Will also strip up to 3 line break/null chars **after** consumed bytes
// Will pack resultant data (attempts to free unused memory), can be called
// with 0 to just pack.
// Returns new length
int ChinkConsume(Chink *ck,int consume){
char *nextc;
int newsize;

    if(consume>0){
        if(consume>ck->len)consume=ck->len;
        // TODO: no comment!
        if(((*(ck->ch+consume)==ck->tr)||(*(ck->ch+consume)==ck->tn)||(*(ck->ch+consume)==0))&&consume<ck->len)consume++;
        if(((*(ck->ch+consume)==ck->tr)||(*(ck->ch+consume)==ck->tn)||(*(ck->ch+consume)==0))&&consume<ck->len)consume++;
        if(((*(ck->ch+consume)==ck->tr)||(*(ck->ch+consume)==ck->tn)||(*(ck->ch+consume)==0))&&consume<ck->len)consume++;
    }

    // TODO: You need two cursors to eliminate superfluous move!
    newsize=ck->len-consume;
    ck->len=newsize;
    memmove(ck->ch,ck->ch+consume,newsize);
    if(newsize<CHINKBLOK)newsize=CHINKBLOK;
    nextc=realloc(ck->ch,newsize);
    if(nextc)ck->ch=nextc;

}

    

