
#include "../sqlite/sqlite3.h"
#include <stdio.h>
#include <string.h>
#ifdef WIN32
    #include <winsock2.h>
    #include  <ws2tcpip.h>
#else
    #include <sys/socket.h>
    #include <netinet/in.h>
    #include <errno.h>
#endif
#include <fcntl.h>
#include "uSQLsec.h"
#include "Chunk.h"
#include "chink.h"
#include "uSQLct.h"
#include "gLog.h"


int callback(void *ct, int argc, char **argv, char **ColName){
int i,ll;
Chink *op;
uSQLct *uct;
char fbuf[256];

    // TODO: Make error and abort if Chink can't extend (eg. Max size)

    uct=(uSQLct*)ct;
	op=uct->CBop;
	ChinkReset(op);

    if(uct->RecNum>uct->MaxRepRecs){
    	ChinkAdd(op,":Err : MAXRECORDS ",-1);
    	ChinkAdd(op,&uct->tr,1);
   		uSQLctPut(uct,op->ch,op->len);
        return -1;
    }

	if(uct->Human){
		ChinkAdd(op,fbuf,sprintf(fbuf,"====( %i )====%c",++uct->RecNum,uct->tr));
	}
	else{
		if(uct->RecNum==0){
			// Put list of headers
			for(i=0;i<argc;i++)
                ChinkAdd(op,fbuf,sprintf(fbuf,":H%i:%i %s%c",i+1,strlen(ColName[i]),ColName[i],uct->tr));
            ChinkAdd(op,fbuf,sprintf(fbuf,":R%c",uct->tr));
		}
        ++uct->RecNum;
	}


	for(i=0; i<argc; i++){
		if(uct->Human){
			ChinkAdd(op,ColName[i],-1);
			ChinkAdd(op," = ",3);
			ChinkAdd(op,(argv[i] ? argv[i] : "NULL"),-1);
			ChinkAdd(op,&uct->tr,1);
		}
		else{
			if(argv[i]){
                // Most results are small!
                if((ll=strlen(argv[i]))<=32)
                    // If result is small and does not start with ':' or '!' then supress :Fx
                    if((*argv[i]==':')||(*argv[i]=='!'))
                        ChinkAdd(op,fbuf,sprintf(fbuf,":F%i %s%c",ll,argv[i],uct->tr));
                    else{
                        ChinkAdd(op,argv[i],strlen(argv[i]));
                        ChinkAdd(op,&uct->tr,1);
                    }
                // This form avoids passing big blobs etc. thruough sprintf
                else{
                	ChinkAdd(op,fbuf,sprintf(fbuf,":F%i ",ll));
                	ChinkAdd(op,argv[i],ll);
                	ChinkAdd(op,&uct->tr,1);
                }
			}
			else
                ChinkAdd(op,fbuf,sprintf(fbuf,"!%c",uct->tr));
		}

	}
	uSQLctPut(uct,op->ch,op->len);
	return 0;
}


// Following creation set sec, parent, netlevel, reclimit & maxbuffer
uSQLct *uSQLctInit(sqlite3 *idb,int unID,int sock,uSQLsec *sec,int netlevel,int maxbuffer,int MaxRetRec,int MaxQlength,int MaxRepRecs){
uSQLct *nct=NULL;
int flgs;
struct linger lg;
socklen_t lgs;
unsigned int nonblock=1;

    nct=(uSQLct*)malloc(sizeof(uSQLct));
    if(!nct)return NULL;
    // santize nct before use
    nct->sec=sec;
    //nct->parent=NULL;
    nct->db=NULL;
    nct->uniqueID=unID;
    nct->chBuff=NULL;
    nct->qbuf=NULL;
    nct->CBop=NULL;
    nct->maxbuffer=maxbuffer;
    nct->MaxRetRec=MaxRetRec;
    nct->MaxQlength=MaxQlength;
    nct->MaxRepRecs=MaxRepRecs;
    nct->netsec=netlevel;
    *(nct->user)=0;
    *(nct->pass)=0;

    nct->socknum=sock;

    // set linger
    lg.l_onoff=1;
    lg.l_linger=20;
    lgs=sizeof(lg);
    setsockopt(nct->socknum,SOL_SOCKET,SO_LINGER,(const char *)&lg,lgs);


    // Set socket options
#ifdef WIN32
    // set socket non-blocking
    ioctlsocket(nct->socknum,FIONBIO,(u_long FAR*)&nonblock);
#else
    flgs=fcntl(nct->socknum,F_GETFL,0);
    if(flgs==-1)goto fail;
    flgs|=O_NONBLOCK;
    fcntl(nct->socknum,F_SETFL,flgs);
#endif

    if(!(nct->chBuff=MakeChunk()))goto fail;

    if(!(nct->CBop=ChinkMake(nct->MaxRetRec)))goto fail;

    nct->db=idb;

    nct->Human=0;
    nct->uselevel=0;
    nct->tr='\r';

    return nct;

fail:
    uSQLctDestroy(nct);
    return NULL;
}

void uSQLctDestroy(uSQLct *ct){
	if(ct->chBuff)ChunkDestroy(ct->chBuff);
	if(ct->qbuf)ChinkDestroy(ct->qbuf);
	if(ct->CBop)ChinkDestroy(ct->CBop);
#ifdef WIN32
    closesocket(ct->socknum);
#else
    close(ct->socknum);
#endif
	free(ct);
}

void TrySend(uSQLct *ct){
int sent;


    while(ChunkLeadSz(ct->chBuff)){
        sent=send(ct->socknum,ChunkLead(ct->chBuff),ChunkLeadSz(ct->chBuff),0);
        if(sent<1)break;
        ChunkConsume(ct->chBuff,sent);
	}
}

void uSQLctPut(uSQLct *ct,char *txt,int len){
int sent,err;

	// If output buffer is empty we may try an immediate send.
	if(ChunkGrossSize(ct->chBuff)==0){
		sent=send(ct->socknum,txt,len,0);
		if(sent==len)return;
		if(sent<0)sent=0;
		txt+=sent;len-=sent;
	}
	// TODO check Chunk is not over maxsize (or limit chunk)
	err=ChunkAdd(ct->chBuff,txt,len);
	if(err)
        fprintf(stderr,"\nMemory error not yet handled\n");
	TrySend(ct);
}



void DoPPragma(uSQLct *ct,char *Err,char *query){
char *cur, npass[70],opass[70];
int pargs,a1,a2;

	sprintf(Err,"%s%c",Err,ct->tr);

	if(!strncmp(query," HUMAN",6)){
		ct->Human=1;
		sprintf(Err,"%s:PPRAGMA HUMAN MODE%c:OK%c",Err,ct->tr,ct->tr);
	}
	else if(!strncmp(query," MACHINE",8)){
		ct->Human=0;
		sprintf(Err,"%s:PPRAGMA MACHINE MODE%c:OK%c",Err,ct->tr,ct->tr);
	}
	else if(!strncmp(query," VER",4)){
		sprintf(Err,"%s:PPRAGMA VER uSQLiteServer 0.4%c:OK%c",Err,ct->tr,ct->tr);
	}	
	else if(!strncmp(query," ID",3)){
		sprintf(Err,"%s:PPRAGMA ID %i%c:OK%c",Err,ct->uniqueID,ct->tr,ct->tr);
	}
	else if(!strncmp(query," ETX",4)){
		ct->qbuf->tr=ct->qbuf->tn=ct->tr=3;
		sprintf(Err,"%s:PPRAGMA ETX MODE %s%c:OK%c",Err,ct->user,ct->tr,ct->tr);
	}
	else if(!strncmp(query," USER ",6)){
		cur=query+=6;
		while(*cur>' ')cur++;
		*cur=0;
		//Limit username length
		strncpy(ct->user,query,32);
		sprintf(Err,"%s:PPRAGMA USER %s%c:OK%c",Err,ct->user,ct->tr,ct->tr);
	}
    else if(!strncmp(query," PASS ",6)){
		cur=query+=6;
		while(*cur>' ')cur++;
		*cur=0;
		strncpy(ct->pass,query,32);
		ct->uselevel=uSQLsecGetUserLevel(ct->sec,ct->user,ct->pass);
		ct->uselevel&=ct->netsec;
		sprintf(Err,"%s:PPRAGMA USELEVEL %i%c:OK%c",Err,ct->uselevel,ct->tr,ct->tr);
	}


	else if(!strncmp(query," EXEC ",6)){
		pargs=sscanf(query+6,"%i %i",&a1,&a2);
		if(pargs!=2)
            sprintf(Err,"%s:Err : PPRAGMA : EXEC args invalid%c:OK%c",Err,ct->tr,ct->tr);
        else{
        	pargs=execDB(ct->db,a1,a2);
        	if(pargs!=(a2+1))
                sprintf(Err,"%s:Err : PPRAGMA : EXEC Failed at %i%c:OK%c",Err,pargs,ct->tr,ct->tr);
            else
                sprintf(Err,"%s:OK%c",Err,ct->tr);
        }
	}

	else if(!strncmp(query," NEWPASS ",9)){
		pargs=sscanf(query+9,"%s %s",npass,opass);
		if(pargs==1){*opass=0;pargs=2;}
		if(pargs!=2){
			sprintf(Err,"%s:Err : PPRAGMA : NEWPASS arguments error (oldpass newpass)%c:OK%c",Err,ct->tr,ct->tr);
		}
		else{
			if((strlen(npass)>32)||(strlen(opass)>32)){
				sprintf(Err,"%s:Err : PPRAGMA : user/pass max 32 chars%c:OK%c",Err,ct->tr,ct->tr);
			}
			else{
				if(uSQLsecNewPass(ct->sec,ct->user,npass,opass)==0)
					sprintf(Err,"%s:Err : PPRAGMA : NEWPASS failed%c:OK%c",Err,ct->tr,ct->tr);
                else
                    sprintf(Err,"%s:PPRAGMA NEWPASS %s%c:OK%c",Err,ct->user,ct->tr,ct->tr);
			}
		}
	}

	else{
		sprintf(Err,":Err : PPRAGMA : Unknown command%c:OK%c",ct->tr,ct->tr);
	}
}

// If I don't use void for ct I get a compile warning, and I have 0 tollerance
// for these. Is there any way to cast without superfluous assignment or 
// casting each use?
int xAuth(void *vct,int op,const char* arg1,const char* arg2,const char* arg3,const char* db){
int access;
uSQLct *ct;

    ct=(uSQLct*)vct;

    switch(op){
        case SQLITE_READ:
	   	case SQLITE_SELECT:
            access=ct->uselevel & 0x01;
            break;
        case SQLITE_UPDATE:
            access=ct->uselevel & 0x02;
            break;
        case SQLITE_INSERT:
            access=ct->uselevel & 0x04;
            break;
        case SQLITE_DELETE:
            access=ct->uselevel & 0x08;
            break;
        default:
            access=ct->uselevel & 0x10;
    }
    
    if(access){
    	return SQLITE_OK;
    }
    else{
    	return SQLITE_DENY;
    }
		
}

void NewQuery(uSQLct *ct,char *query,int len){
int rc;
char *zErrMsg = 0;
char Err[256];

    *Err=0;
	ct->RecNum=0;

	if(!strncmp(query,":PPRAGMA ",9)){
		DoPPragma(ct,Err,query+8);
	}
	else{
		sqlite3_set_authorizer(ct->db,xAuth,ct);
        rc = sqlite3_exec(ct->db, query, callback, ct, &zErrMsg);
        if( rc!=SQLITE_OK ){
            sprintf(Err,":Err : SQL error : %s%c:OK%c",zErrMsg,ct->tr,ct->tr);
        }
        else{
            sprintf(Err,":OK%c",ct->tr);
        }
        sqlite3_free(zErrMsg);
	}
	uSQLctPut(ct,Err,strlen(Err));

}

// Expands HTTP GET Esc codes
// No longer used?
int escGet(char *in, int *len){
char *cur,*end;
int val;

    cur=in;
    end=in+(*len);

    while(cur<end){
        if(*cur=='%'){
            if(sscanf(cur+1,"%2X",&val)!=1)return -1;
            *in=val;
            in++;
            cur+=3;
            len-=2;
        }
        else
        *in++=*cur++;
    }
    return 0;
}


int uSQLctPoll(uSQLct *ct){
int rxed,ll;
Chink *qbuf;
char achar;

    qbuf=ct->qbuf;
    // Create Chink on first invocation
    if(!qbuf){qbuf=ChinkMake(ct->MaxQlength);
        if(!qbuf){
            //Log("uSQLct","Unable to make Chink buffer",0);
            return -2; // Fatal error
        }
        else{
            ct->qbuf=qbuf;
        }
    }

	// We do not even parse input until reply buffer
	// is empty.
	if(ChunkGrossSize(ct->chBuff)){
		TrySend(ct);
		// If there is still data to be sent then return -1
		// This means the server will select on output
		if(ChunkGrossSize(ct->chBuff))return -1;
	}

	else{
        // With no data to go we may try parsing input

        
        do{
            if(ChinkBuffFor(qbuf,2048)<0)break;
            
            rxed=recv(ct->socknum,ChinkIns(qbuf),2048,0);
 
        if(rxed>0)qbuf->len+=rxed;
        }while(rxed>0);
                    

        if((ll=ChinkBreak(qbuf))>0){
            NewQuery(ct,qbuf->ch,11);
            ChinkConsume(qbuf,ll);
            return 1;
        }
        return 0;
	}
}

// Verify that the socket connection is still valid
// will return 0 if no problem and in wait
// If data to Tx will return quantity
// Returns negative if socket failed
// TODO: could we benefit from more verification here?
int uSQLctVer(uSQLct *ct){
int stat;
char foo;


    // recv will return n chars or 'wouldblock'. Only when disconecceted
    // returns 0
    if((stat=recv(ct->socknum,&foo,1,MSG_PEEK))<1){
    	if(!stat)return -1;
    	#ifdef WIN32
    	stat=WSAGetLastError();
    	if(stat!=WSAEWOULDBLOCK)return -1;
    	#else
    	if(errno!=EWOULDBLOCK)return -1;
    	#endif
    }

    return ((stat=ChunkGrossSize(ct->chBuff))>0)?stat:0;

}




