#include <stdlib.h>
#include <string.h>
#include "sha2.h"
#include "../sqlite/sqlite3.h"
#include "uSQLsec.h"
#include "gLog.h"

extern pLog Log;


uSQLsec *uSQLsecOpen(char *secdb){
uSQLsec *nsec;
int err;

    nsec=malloc(sizeof(uSQLsec));
    err=sqlite3_open(secdb,&nsec->db);
    if(err){
    	Log("uSQLsec","Sec db open error",0);
    	sqlite3_close(nsec->db);
    	free(nsec);
    	return NULL;
    }

    Log("uSQLsec","Sec db opened",0);
    return nsec;
}

void uSQLsecDestroy(uSQLsec *sec){
	sqlite3_close(sec->db);
	free(sec);
}
	
		
int uSQLsecGetNetLevel(uSQLsec *sec,char *ip){
char query[]={"Select Address,Mask,Level from Networks"};
char **res;
int rows,cols,level=0;
unsigned int imask,ipval,lastmask=0;

	sqlite3_get_table(sec->db,query,&res,&rows,&cols,NULL);
	if(cols!=3)goto except;
	
	/* Each line of the result will be compared with the remote
	 ip masked by the the mask for that row. If multiple valid
	 entries are found (different levels of mask) then the level used
	 will be that corresponding to the highest mask value (that is the
	 most explicit entry for the ip */
	 
	ipval=uSQLsecIP2int(ip); 
	while(rows){
		imask=uSQLsecIP2int(res[(rows*3)+1]);
		// Only consider if match is better than a previous match
		if(imask>lastmask){
			// Match masked addresses
			if(!((uSQLsecIP2int(res[rows*3])^ipval)& imask)){
				level=atoi(res[(rows*3)+2]);
				lastmask=imask;
			}
		}
		rows--;
	}
	
except:
	sqlite3_free_table(res);
	return level;
}
	


int uSQLsecGetUserLevel(uSQLsec *sec,char *user,char *pass){
char query[256],hash[64];
char **res;
int rows,cols,level=0;

    if((strlen(user)>32)||(strlen(pass)>32))
            return 0;

    *hash=0;
	if(strlen(pass)>0)uSQLsecDoSHA2(hash,"uSQLsecSalt",user,pass);
	sprintf(query,"Select Level from Users where UserName='%s' and Password='%s'",user,hash);
		
	sqlite3_get_table(sec->db,query,&res,&rows,&cols,NULL);
	if((rows==1) && (!strcmp(res[0],"Level")))level=atoi(res[1]);
		
	sqlite3_free_table(res);
	return level;
	
}

void uSQLsecDoSHA2(char *res, char *d1,char *d2, char *d3){
sha224_ctx mctx;
unsigned char dg[32],fbuf[4],*cur; 

	memset(dg,7,32);
	sha224_init(&mctx);
	if(d1)sha224_update(&mctx,(unsigned char *) d1, strlen(d1));
	if(d2)sha224_update(&mctx,(unsigned char *) d2, strlen(d2));
	if(d3)sha224_update(&mctx,(unsigned char *) d3, strlen(d3));
	sha224_final(&mctx, dg);
	
	*res=0;
	for(cur=dg;cur<dg+28;cur++){
		sprintf(fbuf,"%02X",(int)*cur);
		strcat(res,fbuf);
	}	
}

int uSQLsecNewPass(uSQLsec *sec,char *user,char *newpass,char *oldpass){
char query[256],npass[64],opass[64];
char **res;
int rows,cols,level=0,ret,changes;
char *Err;

    if((strlen(user)>32)||(strlen(newpass)>32)||(strlen(oldpass)>32))
            return 0;

	if(strlen(newpass)<1)return 0;
	*npass=0;
	uSQLsecDoSHA2(npass,"uSQLsecSalt",user,newpass);
	
	*opass=0;
	if(strlen(oldpass)>0)uSQLsecDoSHA2(opass,"uSQLsecSalt",user,oldpass);
	
	sprintf(query,"Update users set Password='%s' where Username='%s' and Password='%s'",npass,user,opass);
	ret=sqlite3_exec(sec->db,query,NULL,NULL,&Err);
	changes=sqlite3_changes(sec->db);
	
	return (ret==SQLITE_OK) && (changes==1);	
	
}
	


unsigned int uSQLsecIP2int(char *ip){
unsigned int flds[4];
	if(sscanf(ip,"%u.%u.%u.%u",flds,flds+1,flds+2,flds+3)!=4)return 0;
	return (flds[0]*0x1000000)+(flds[1]*0x10000)+(flds[2]*0x100)+flds[3];
}

	
